#include "apple-orbit-library.h"
#include "examples-toolkit.h"
#include <signal.h>

static void
server_shutdown (int sig)
{
	extern CORBA_ORB global_orb;
	CORBA_Environment  local_ev[1];
	CORBA_exception_init(local_ev);

	if (global_orb != CORBA_OBJECT_NIL)
	{
		CORBA_ORB_shutdown (global_orb, FALSE, local_ev);
		etk_abort_if_exception (local_ev, "caught exception");
	}
}

PortableServer_POA
server_create_multi_threaded_poa (CORBA_ORB  orb,
				  PortableServer_POA         poa,
				  PortableServer_POAManager  poa_mgr,
				  const gchar                *name,
				  CORBA_Environment          *ev)
{
	gint MAX_POLICIES = 1;
	PortableServer_POA  child_poa = CORBA_OBJECT_NIL;
	CORBA_PolicyList *poa_policies;

	poa_policies = CORBA_PolicyList__alloc ();
	poa_policies->_maximum = MAX_POLICIES;
	poa_policies->_length  = MAX_POLICIES;
	poa_policies->_buffer  = CORBA_PolicyList_allocbuf (MAX_POLICIES);
	CORBA_sequence_set_release (poa_policies, CORBA_TRUE);
	poa_policies->_buffer[0] = (CORBA_Policy)
	PortableServer_POA_create_thread_policy 
			(poa, PortableServer_ORB_CTRL_MODEL, ev);

	child_poa = PortableServer_POA_create_POA 
			(poa, name, poa_mgr, poa_policies, ev);
	if (etk_raised_exception(ev)) 
		goto failed_create_poa;

	ORBit_ObjectAdaptor_set_thread_hint ((ORBit_ObjectAdaptor) child_poa, 
						ORBIT_THREAD_HINT_PER_REQUEST);

	CORBA_Policy_destroy (poa_policies->_buffer[0], ev); 
	if (etk_raised_exception(ev)) 
		goto failed;
	CORBA_free (poa_policies);
	return child_poa;

failed_create_poa:
	/* FIXME, in case of error, ev is set, but destructor should not
	 * return except anyway */
	CORBA_Policy_destroy (poa_policies->_buffer[0], ev); 
	CORBA_free (poa_policies);
failed:
	return CORBA_OBJECT_NIL;
}


void 
server_init (int               *argc_ptr, 
			 char              *argv[],
			 CORBA_ORB         *orb,
			 CORBA_Environment *ev)
{
	extern PortableServer_POA rootpoa;
	extern PortableServer_POAManager rootpoa_mgr; 

	CORBA_Environment  local_ev[1];
	CORBA_exception_init(local_ev);

	/* init signal handling */
	signal(SIGTERM, server_shutdown);

	/* create Object Request Broker (ORB) */
	
	(*orb) = CORBA_ORB_init(argc_ptr, argv, "orbit-local-mt-orb", ev);
	if (etk_raised_exception(ev)) 
		goto failed_orb;

	rootpoa = (PortableServer_POA) 
	CORBA_ORB_resolve_initial_references(*orb, "RootPOA", ev);
	if (etk_raised_exception(ev)) 
		goto failed_poa;

	rootpoa_mgr = PortableServer_POA__get_the_POAManager(rootpoa, ev);
	if (etk_raised_exception(ev)) 
		goto failed_poamanager;

	PortableServer_POAManager_activate(rootpoa_mgr, ev);
	if (etk_raised_exception(ev)) 
		goto failed_activation;

	CORBA_Object_release ((CORBA_Object) rootpoa_mgr, ev);
	return;

failed_activation:
failed_poamanager:
	CORBA_Object_release ((CORBA_Object) rootpoa_mgr, local_ev);
failed_poa:
	CORBA_ORB_destroy(*orb, local_ev);		
failed_orb:
	return;
}


void 
server_run (CORBA_ORB          orb,
			CORBA_Environment *ev)
{
	/* enter main loop until SIGINT or SIGTERM */
	
	CORBA_ORB_run(orb, ev);
	if (etk_raised_exception(ev)) return;

	/* user pressed SIGINT or SIGTERM and in signal handler
	 * CORBA_ORB_shutdown(.) has been called */
}

/* Releases @servant object and finally destroys @orb. If error
 * occures @ev points to exception object on return.
 */
void 
server_cleanup (CORBA_ORB   orb,
		PortableServer_POA  poa,
		CORBA_Object        ref,
		CORBA_Environment  *ev)
{
	PortableServer_ObjectId   *objid       = NULL;

	objid = PortableServer_POA_reference_to_id (poa, ref, ev);
	if (etk_raised_exception(ev)) return;
		
	/* Servant: deactivatoin - will invoke  __fini destructor */
	PortableServer_POA_deactivate_object (poa, objid, ev);
	if (etk_raised_exception(ev)) return;

	PortableServer_POA_destroy (poa, TRUE, FALSE, ev);
	if (etk_raised_exception(ev)) return;

	CORBA_free (objid);

	CORBA_Object_release ((CORBA_Object) poa, ev);
	if (etk_raised_exception(ev)) return;
	CORBA_Object_release (ref, ev);
	if (etk_raised_exception(ev)) return;
	/* ORB: tear down the ORB */
	if (orb != CORBA_OBJECT_NIL)
	{
		/* going to destroy orb.. */
		CORBA_ORB_destroy(orb, ev);
		if (etk_raised_exception(ev)) return;
	}
}
